from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.layers.preprocessing.preprocessing_layer import (
    PreprocessingLayer,
)
from keras_hub.src.utils.preset_utils import builtin_presets
from keras_hub.src.utils.preset_utils import find_subclass
from keras_hub.src.utils.preset_utils import get_preset_loader
from keras_hub.src.utils.preset_utils import get_preset_saver
from keras_hub.src.utils.python_utils import classproperty


@keras_hub_export("keras_hub.layers.AudioConverter")
class AudioConverter(PreprocessingLayer):
    """Convert raw audio for models that support audio input.

    This class converts from raw audio tensors of any length, to preprocessed
    audio for pretrained model inputs. It is meant to be a convenient way to
    write custom preprocessing code that is not model specific. This layer
    should be instantiated via the `from_preset()` constructor, which will
    create the correct subclass of this layer for the model preset.

    The layer will take as input a raw audio tensor with shape `(batch_size,
    num_samples)`, and output a preprocessed audio input for modeling. The exact
    structure of the preprocessed input will vary per model. Preprocessing
    will often include computing a spectogram of the raw audio signal.

    Examples:
    ```python
    # Load an audio converter from a preset.
    converter = keras_hub.layers.AudioConverter.from_preset("whisper_base_en")
    # Convert some raw audio input.
    converter(np.ones(2, 1_000))
    ```
    """

    backbone_cls = None

    def audio_shape(self):
        """Returns the preprocessed size of a single audio sample."""
        return (None,)

    @classproperty
    def presets(cls):
        """List built-in presets for an `AudioConverter` subclass."""
        return builtin_presets(cls)

    @classmethod
    def from_preset(
        cls,
        preset,
        **kwargs,
    ):
        """Instantiate a `keras_hub.layers.AudioConverter` from a model preset.

        A preset is a directory of configs, weights and other file assets used
        to save and load a pre-trained model. The `preset` can be passed as
        one of:

        1. a built-in preset identifier like `'whisper_base_en'`
        2. a Kaggle Models handle like
           `'kaggle://user/whisper/keras/whisper_base_en'`
        3. a Hugging Face handle like `'hf://user/whisper_base_en'`
        4. a path to a local preset directory like `'./whisper_base_en'`

        You can run `cls.presets.keys()` to list all built-in presets available
        on the class.

        This constructor can be called in one of two ways. Either from the base
        class like `keras_hub.models.AudioConverter.from_preset()`, or from a
        model class like `keras_hub.models.WhisperAudioConverter.from_preset()`.
        If calling from the base class, the subclass of the returning object
        will be inferred from the config in the preset directory.

        Args:
            preset: string. A built-in preset identifier, a Kaggle Models
                handle, a Hugging Face handle, or a path to a local directory.
            load_weights: bool. If `True`, the weights will be loaded into the
                model architecture. If `False`, the weights will be randomly
                initialized.

        Examples:
        ```python
        # Load an audio converter from a preset.
        converter = keras_hub.layers.AudioConverter.from_preset(
            "whisper_base_en"
        )
        # Convert some raw mono channel audio input.
        converter(np.ones(2, 1_000))
        ```
        """
        loader = get_preset_loader(preset)
        backbone_cls = loader.check_backbone_class()
        if cls.backbone_cls != backbone_cls:
            cls = find_subclass(preset, cls, backbone_cls)
        return loader.load_audio_converter(cls, **kwargs)

    def save_to_preset(self, preset_dir):
        """Save audio converter to a preset directory.

        Args:
            preset_dir: The path to the local model preset directory.
        """
        saver = get_preset_saver(preset_dir)
        saver.save_audio_converter(self)



from keras_hub.src.models.efficientnet.efficientnet_backbone import (
    EfficientNetBackbone,
)
from keras_hub.src.models.efficientnet.efficientnet_presets import (
    backbone_presets,
)
from keras_hub.src.utils.preset_utils import register_presets

register_presets(backbone_presets, EfficientNetBackbone)



import sys

import keras
from absl import logging

try:
    import tensorflow as tf
except ImportError:
    tf = None


def clone_initializer(initializer):
    """Clones an initializer to ensure a new seed.

    As of tensorflow 2.10, we need to clone user passed initializers when
    invoking them twice to avoid creating the same randomized initialization.
    """
    # If we get a string or dict, just return as we cannot and should not clone.
    if not isinstance(initializer, keras.initializers.Initializer):
        return initializer
    config = initializer.get_config()
    return initializer.__class__.from_config(config)


def print_msg(message, line_break=True):
    """Print the message to absl logging or stdout."""
    # Copied from core Keras.
    if keras.utils.is_interactive_logging_enabled():
        if line_break:
            sys.stdout.write(message + "\n")
        else:
            sys.stdout.write(message)
        sys.stdout.flush()
    else:
        logging.info(message)


# Register twice for backwards compat.
@keras.saving.register_keras_serializable(package="keras_hub")
@keras.saving.register_keras_serializable(package="keras_nlp")
def gelu_approximate(x):
    return keras.activations.gelu(x, approximate=True)


def standardize_data_format(data_format):
    if data_format is None:
        return keras.config.image_data_format()
    data_format = str(data_format).lower()
    if data_format not in {"channels_first", "channels_last"}:
        raise ValueError(
            "The `data_format` argument must be one of "
            "{'channels_first', 'channels_last'}. "
            f"Received: data_format={data_format}"
        )
    return data_format


def has_flash_attention_support():
    if (
        hasattr(keras.config, "is_flash_attention_enabled")
        and keras.config.backend() == "jax"
    ):
        try:
            from jax.nn import dot_product_attention as dot_product_attention
        except ImportError:
            logging.warning(
                "Flash attention is not supported in your current JAX version. "
                "Please update it by following the official guide: "
                "https://jax.readthedocs.io/en/latest/installation.html"
            )
            return False
        return True
    else:
        return False


def running_on_tpu():
    backend = keras.config.backend()
    if backend == "jax":
        import jax

        devices = jax.devices()
        return any(d.platform == "tpu" for d in devices)
    elif backend == "tensorflow":
        import tensorflow as tf

        return bool(tf.config.list_logical_devices("TPU"))
    elif backend == "torch":
        return False


def running_on_gpu():
    backend = keras.config.backend()
    if backend == "jax":
        import jax

        devices = jax.devices()
        return any(d.platform == "gpu" for d in devices)
    elif backend == "tensorflow":
        import tensorflow as tf

        return bool(tf.config.list_logical_devices("GPU"))
    elif backend == "torch":
        import torch

        return torch.cuda.is_available()



import keras
from keras import layers

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.feature_pyramid_backbone import FeaturePyramidBackbone


@keras_hub_export("keras_hub.models.CSPDarkNetBackbone")
class CSPDarkNetBackbone(FeaturePyramidBackbone):
    """This class represents Keras Backbone of CSPDarkNet model.

    This class implements a CSPDarkNet backbone as described in
    [CSPNet: A New Backbone that can Enhance Learning Capability of CNN](
        https://arxiv.org/abs/1911.11929).

    Args:
        stackwise_num_filters:  A list of ints, filter size for each dark
            level in the model.
        stackwise_depth: A list of ints, the depth for each dark level in the
            model.
        block_type: str. One of `"basic_block"` or `"depthwise_block"`.
            Use `"depthwise_block"` for depthwise conv block
            `"basic_block"` for basic conv block.
            Defaults to "basic_block".
        image_shape: tuple. The input shape without the batch size.
            Defaults to `(None, None, 3)`.

    Examples:
    ```python
    input_data = np.ones(shape=(8, 224, 224, 3))

    # Pretrained backbone
    model = keras_hub.models.CSPDarkNetBackbone.from_preset(
        "csp_darknet_tiny_imagenet"
    )
    model(input_data)

    # Randomly initialized backbone with a custom config
    model = keras_hub.models.CSPDarkNetBackbone(
        stackwise_num_filters=[128, 256, 512, 1024],
        stackwise_depth=[3, 9, 9, 3],
    )
    model(input_data)
    ```
    """

    def __init__(
        self,
        stackwise_num_filters,
        stackwise_depth,
        block_type="basic_block",
        image_shape=(None, None, 3),
        **kwargs,
    ):
        # === Functional Model ===
        channel_axis = (
            -1 if keras.config.image_data_format() == "channels_last" else 1
        )
        apply_ConvBlock = (
            apply_darknet_conv_block_depthwise
            if block_type == "depthwise_block"
            else apply_darknet_conv_block
        )
        base_channels = stackwise_num_filters[0] // 2

        image_input = layers.Input(shape=image_shape)
        x = image_input  # Intermediate result.
        x = apply_focus(channel_axis, name="stem_focus")(x)
        x = apply_darknet_conv_block(
            base_channels,
            channel_axis,
            kernel_size=3,
            strides=1,
            name="stem_conv",
        )(x)

        pyramid_outputs = {}
        for index, (channels, depth) in enumerate(
            zip(stackwise_num_filters, stackwise_depth)
        ):
            x = apply_ConvBlock(
                channels,
                channel_axis,
                kernel_size=3,
                strides=2,
                name=f"dark{index + 2}_conv",
            )(x)

            if index == len(stackwise_depth) - 1:
                x = apply_spatial_pyramid_pooling_bottleneck(
                    channels,
                    channel_axis,
                    hidden_filters=channels // 2,
                    name=f"dark{index + 2}_spp",
                )(x)

            x = apply_cross_stage_partial(
                channels,
                channel_axis,
                num_bottlenecks=depth,
                block_type="basic_block",
                residual=(index != len(stackwise_depth) - 1),
                name=f"dark{index + 2}_csp",
            )(x)
            pyramid_outputs[f"P{index + 2}"] = x

        super().__init__(inputs=image_input, outputs=x, **kwargs)

        # === Config ===
        self.stackwise_num_filters = stackwise_num_filters
        self.stackwise_depth = stackwise_depth
        self.block_type = block_type
        self.image_shape = image_shape
        self.pyramid_outputs = pyramid_outputs

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "stackwise_num_filters": self.stackwise_num_filters,
                "stackwise_depth": self.stackwise_depth,
                "block_type": self.block_type,
                "image_shape": self.image_shape,
            }
        )
        return config


def apply_focus(channel_axis, name=None):
    """A block used in CSPDarknet to focus information into channels of the
    image.

    If the dimensions of a batch input is (batch_size, width, height, channels),
    this layer converts the image into size (batch_size, width/2, height/2,
    4*channels). See [the original discussion on YoloV5 Focus Layer](https://github.com/ultralytics/yolov5/discussions/3181).

    Args:
        name: the name for the lambda layer used in the block.

    Returns:
        a function that takes an input Tensor representing a Focus layer.
    """

    def apply(x):
        return layers.Concatenate(axis=channel_axis, name=name)(
            [
                x[..., ::2, ::2, :],
                x[..., 1::2, ::2, :],
                x[..., ::2, 1::2, :],
                x[..., 1::2, 1::2, :],
            ],
        )

    return apply


def apply_darknet_conv_block(
    filters,
    channel_axis,
    kernel_size,
    strides,
    use_bias=False,
    activation="silu",
    name=None,
):
    """
    The basic conv block used in Darknet. Applies Conv2D followed by a
    BatchNorm.

    Args:
        filters: Integer, the dimensionality of the output space (i.e. the
            number of output filters in the convolution).
        kernel_size: An integer or tuple/list of 2 integers, specifying the
            height and width of the 2D convolution window. Can be a single
            integer to specify the same value both dimensions.
        strides: An integer or tuple/list of 2 integers, specifying the strides
            of the convolution along the height and width. Can be a single
            integer to the same value both dimensions.
        use_bias: Boolean, whether the layer uses a bias vector.
        activation: the activation applied after the BatchNorm layer. One of
            "silu", "relu" or "leaky_relu", defaults to "silu".
        name: the prefix for the layer names used in the block.
    """
    if name is None:
        name = f"conv_block{keras.backend.get_uid('conv_block')}"

    def apply(inputs):
        x = layers.Conv2D(
            filters,
            kernel_size,
            strides,
            padding="same",
            data_format=keras.config.image_data_format(),
            use_bias=use_bias,
            name=name + "_conv",
        )(inputs)

        x = layers.BatchNormalization(axis=channel_axis, name=name + "_bn")(x)

        if activation == "silu":
            x = layers.Lambda(lambda x: keras.activations.silu(x))(x)
        elif activation == "relu":
            x = layers.ReLU()(x)
        elif activation == "leaky_relu":
            x = layers.LeakyReLU(0.1)(x)

        return x

    return apply


def apply_darknet_conv_block_depthwise(
    filters, channel_axis, kernel_size, strides, activation="silu", name=None
):
    """
    The depthwise conv block used in CSPDarknet.

    Args:
        filters: Integer, the dimensionality of the output space (i.e. the
            number of output filters in the final convolution).
        kernel_size: An integer or tuple/list of 2 integers, specifying the
            height and width of the 2D convolution window. Can be a single
            integer to specify the same value both dimensions.
        strides: An integer or tuple/list of 2 integers, specifying the strides
            of the convolution along the height and width. Can be a single
            integer to the same value both dimensions.
        activation: the activation applied after the final layer. One of "silu",
            "relu" or "leaky_relu", defaults to "silu".
        name: the prefix for the layer names used in the block.

    """
    if name is None:
        name = f"conv_block{keras.backend.get_uid('conv_block')}"

    def apply(inputs):
        x = layers.DepthwiseConv2D(
            kernel_size,
            strides,
            padding="same",
            data_format=keras.config.image_data_format(),
            use_bias=False,
        )(inputs)
        x = layers.BatchNormalization(axis=channel_axis)(x)

        if activation == "silu":
            x = layers.Lambda(lambda x: keras.activations.swish(x))(x)
        elif activation == "relu":
            x = layers.ReLU()(x)
        elif activation == "leaky_relu":
            x = layers.LeakyReLU(0.1)(x)

        x = apply_darknet_conv_block(
            filters,
            channel_axis,
            kernel_size=1,
            strides=1,
            activation=activation,
        )(x)

        return x

    return apply


def apply_spatial_pyramid_pooling_bottleneck(
    filters,
    channel_axis,
    hidden_filters=None,
    kernel_sizes=(5, 9, 13),
    activation="silu",
    name=None,
):
    """
    Spatial pyramid pooling layer used in YOLOv3-SPP

    Args:
        filters: Integer, the dimensionality of the output spaces (i.e. the
            number of output filters in used the blocks).
        hidden_filters: Integer, the dimensionality of the intermediate
            bottleneck space (i.e. the number of output filters in the
            bottleneck convolution). If None, it will be equal to filters.
            Defaults to None.
        kernel_sizes: A list or tuple representing all the pool sizes used for
            the pooling layers, defaults to (5, 9, 13).
        activation: Activation for the conv layers, defaults to "silu".
        name: the prefix for the layer names used in the block.

    Returns:
        a function that takes an input Tensor representing an
        SpatialPyramidPoolingBottleneck.
    """
    if name is None:
        name = f"spp{keras.backend.get_uid('spp')}"

    if hidden_filters is None:
        hidden_filters = filters

    def apply(x):
        x = apply_darknet_conv_block(
            hidden_filters,
            channel_axis,
            kernel_size=1,
            strides=1,
            activation=activation,
            name=f"{name}_conv1",
        )(x)
        x = [x]

        for kernel_size in kernel_sizes:
            x.append(
                layers.MaxPooling2D(
                    kernel_size,
                    strides=1,
                    padding="same",
                    data_format=keras.config.image_data_format(),
                    name=f"{name}_maxpool_{kernel_size}",
                )(x[0])
            )

        x = layers.Concatenate(axis=channel_axis, name=f"{name}_concat")(x)
        x = apply_darknet_conv_block(
            filters,
            channel_axis,
            kernel_size=1,
            strides=1,
            activation=activation,
            name=f"{name}_conv2",
        )(x)

        return x

    return apply


def apply_cross_stage_partial(
    filters,
    channel_axis,
    num_bottlenecks,
    residual=True,
    block_type="basic_block",
    activation="silu",
    name=None,
):
    """A block used in Cross Stage Partial Darknet.

    Args:
        filters: Integer, the dimensionality of the output space (i.e. the
            number of output filters in the final convolution).
        num_bottlenecks: an integer representing the number of blocks added in
            the layer bottleneck.
        residual: a boolean representing whether the value tensor before the
            bottleneck should be added to the output of the bottleneck as a
            residual, defaults to True.
        block_type: str. One of `"basic_block"` or `"depthwise_block"`.
            Use `"depthwise_block"` for depthwise conv block
            `"basic_block"` for basic conv block.
            Defaults to "basic_block".
        activation: the activation applied after the final layer. One of "silu",
            "relu" or "leaky_relu", defaults to "silu".
    """

    if name is None:
        uid = keras.backend.get_uid("cross_stage_partial")
        name = f"cross_stage_partial_{uid}"

    def apply(inputs):
        hidden_channels = filters // 2
        ConvBlock = (
            apply_darknet_conv_block_depthwise
            if block_type == "basic_block"
            else apply_darknet_conv_block
        )

        x1 = apply_darknet_conv_block(
            hidden_channels,
            channel_axis,
            kernel_size=1,
            strides=1,
            activation=activation,
            name=f"{name}_conv1",
        )(inputs)

        x2 = apply_darknet_conv_block(
            hidden_channels,
            channel_axis,
            kernel_size=1,
            strides=1,
            activation=activation,
            name=f"{name}_conv2",
        )(inputs)

        for i in range(num_bottlenecks):
            residual_x = x1
            x1 = apply_darknet_conv_block(
                hidden_channels,
                channel_axis,
                kernel_size=1,
                strides=1,
                activation=activation,
                name=f"{name}_bottleneck_{i}_conv1",
            )(x1)
            x1 = ConvBlock(
                hidden_channels,
                channel_axis,
                kernel_size=3,
                strides=1,
                activation=activation,
                name=f"{name}_bottleneck_{i}_conv2",
            )(x1)
            if residual:
                x1 = layers.Add(name=f"{name}_bottleneck_{i}_add")(
                    [residual_x, x1]
                )

        x = layers.Concatenate(name=f"{name}_concat")([x1, x2])
        x = apply_darknet_conv_block(
            filters,
            channel_axis,
            kernel_size=1,
            strides=1,
            activation=activation,
            name=f"{name}_conv3",
        )(x)

        return x

    return apply



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras_hub.api import layers
from keras_hub.api import metrics
from keras_hub.api import models
from keras_hub.api import samplers
from keras_hub.api import tokenizers
from keras_hub.api import utils
from keras_hub.src.utils.preset_utils import upload_preset
from keras_hub.src.version_utils import __version__
from keras_hub.src.version_utils import version
